#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<getopt.h>

#include"mcd.h"
#include"util.h"
#include"word_buffer.h"
#include"json_parser.h"

typedef struct {
  int help;
  int verbose;
  int debug_mode;
  char *program_name;
  char *conll_filename;
  char *mcf_filename;
  char *mcd_filename;
  mcd *mcd_struct;
  char *json_filename;
} context;

void json2mcf_context_free(context *ctx)
{
  if(ctx){
    if(ctx->program_name)
      free(ctx->program_name);
    if(ctx->conll_filename)
      free(ctx->conll_filename);
    if(ctx->mcf_filename)
      free(ctx->mcf_filename);
    if(ctx->mcd_filename)
      free(ctx->mcd_filename);
    if(ctx->mcd_struct)
      mcd_free(ctx->mcd_struct);
    free(ctx);
  }
}

context *json2mcf_context_new(void)
{
  context *ctx = (context *)memalloc(sizeof(context));

  ctx->help = 0;
  ctx->verbose = 0;
  ctx->debug_mode = 0;
  ctx->program_name = NULL;
  ctx->conll_filename = NULL;
  ctx->mcf_filename = NULL;
  ctx->mcd_filename = NULL;
  ctx->mcd_struct = NULL;
  ctx->json_filename = NULL;
  return ctx;
}

void json2mcf_context_general_help_message(context *ctx)
{
  fprintf(stderr, "usage: %s [options]\n", ctx->program_name);
  fprintf(stderr, "Options:\n");
  fprintf(stderr, "\t-h --help             : print this message\n");
  fprintf(stderr, "\t-v --verbose          : activate verbose mode\n");
  fprintf(stderr, "\t-C --mcd              : mcd filename\n");
  fprintf(stderr, "\t-i --mcf              : mcf filename (read from stdin if absent)\n");
  fprintf(stderr, "\t-j --json             : json filename\n");
}

void json2mcf_check_options(context *ctx){
  if(ctx->help){
    json2mcf_context_general_help_message(ctx);
    exit(1);
  }
}

context *json2mcf_context_read_options(int argc, char *argv[])
{
  int c;
  int option_index = 0;
  context *ctx = json2mcf_context_new();

  ctx->program_name = strdup(argv[0]);

  static struct option long_options[7] =
    {
      {"help",                no_argument,       0, 'h'},
      {"verbose",             no_argument,       0, 'v'},
      {"debug",               no_argument,       0, 'd'},
      {"conll",               required_argument, 0, 'o'},
      {"mcd",                 required_argument, 0, 'C'}, 
      {"mcf",                 required_argument, 0, 'i'},
      {"json",                 required_argument, 0, 'j'},
    };
  optind = 0;
  opterr = 0;
  
  while ((c = getopt_long (argc, argv, "hvdo:C:i:j:", long_options, &option_index)) != -1){ 
    switch (c)
      {
      case 'd':
	ctx->debug_mode = 1;
	break;
      case 'h':
	ctx->help = 1;
	break;
      case 'v':
	ctx->verbose = 1;
	break;
      case 'o':
	ctx->conll_filename = strdup(optarg);
	break;
      case 'i':
	ctx->mcf_filename = strdup(optarg);
	break;
      case 'C':
	ctx->mcd_filename = strdup(optarg);
	break;
      case 'j':
	ctx->json_filename = strdup(optarg);
	break;
      }
  }

  if(ctx->mcd_filename){
    ctx->mcd_struct = mcd_read(ctx->mcd_filename, ctx->verbose);
  }
  else{
    ctx->mcd_struct = mcd_build_wpmlgfs();
  }

  return ctx;
}

void update_segment(word_buffer *wb, int start, int end, char *label, char *status_seg, char *status_lab)
{

}

void process_segment(json_attr_val *avl, word_buffer *wb)
{
  int start, end;
  char *label, *status_seg, *status_lab;
  json_attr_val *av;

  for(av = avl; av != NULL; av = av->next){
    //      printf("attr = %s\n", av->attr);
    if(!strcmp(av->attr, "start")){start = (int)(av->val->u.number); continue;}
    if(!strcmp(av->attr, "end")){end = (int)(av->val->u.number); continue;}
    if(!strcmp(av->attr, "label")){label = av->val->u.string; continue;}
    if(!strcmp(av->attr, "status_seg")){status_seg = av->val->u.string; continue;}
    if(!strcmp(av->attr, "status_lab")){status_lab = av->val->u.string; continue;}
  }
  update_segment(wb, start, end, label, status_seg, status_lab);
  //  printf("segment : start = %d end = %d label = %s status_seg = %s status_lab = %s\n", start, end, label, status_seg, status_lab);

}

void process_segments(json_struct *segments, word_buffer *wb)
{
  json_struct *segment;
  printf("process_segments\n");
  for(segment = segments->u.first; segment != NULL; segment = segment->next){
    process_segment(segment->u.attr_val_list, wb);
  }
}

// {"orig": 1, "dest":2, "label": "suj", "status_link": "", "status_lab": "", "timestamp": "", "author": "", "target": ""},


void process_link(json_attr_val *avl, word_buffer *wb)
{
  int orig, dest;
  char *label, *status_link, *status_lab;
  json_attr_val *av;

  for(av = avl; av != NULL; av = av->next){
    //      printf("attr = %s\n", av->attr);
    if(!strcmp(av->attr, "orig")){orig = (int)(av->val->u.number); continue;}
    if(!strcmp(av->attr, "dest")){dest = (int)(av->val->u.number); continue;}
    if(!strcmp(av->attr, "label")){label = av->val->u.string; continue;}
    if(!strcmp(av->attr, "status_link")){status_link = av->val->u.string; continue;}
    if(!strcmp(av->attr, "status_lab")){status_lab = av->val->u.string; continue;}
  }
  //  printf("link : orig = %d dest = %d label = %s status_link = %s status_lab = %s\n", orig, dest, label, status_link, status_lab);
}

void process_links(json_struct *segments, word_buffer *wb)
{
  json_struct *link;
  printf("process_links\n");
  for(link = segments->u.first; link != NULL; link = link->next){
    process_link(link->u.attr_val_list, wb);
  }

}

void process_document(json_struct *document, word_buffer *wb)
{
  json_attr_val *avl = NULL;
  printf("process_document\n");
  for(avl = document->u.attr_val_list; avl != NULL; avl = avl->next){
    if(!strcmp(avl->attr, (char *)"id")) printf("id = %s\n", avl->val->u.string);
    if(!strcmp(avl->attr, (char *)"segments")) process_segments(avl->val, wb);
    if(!strcmp(avl->attr, (char *)"links")) process_links(avl->val, wb);
  }
}

void process_documents(json_struct *documents, word_buffer *wb)
{
  json_struct *document;
  printf("process_documents\n");
  for(document = documents->u.first; document != NULL; document = document->next){
    process_document(document, wb);
  }
}

int main(int argc, char *argv[])
{
  FILE *output_file;
  context *ctx = json2mcf_context_read_options(argc, argv);
  word_buffer *wb = NULL;
  word *w = NULL;
  int first_sentence = 1;
  int new_sentence = 1;
  int index_first_word;
  int index_last_word;
  int sentence_nb = 0;
  json_struct *root = NULL;
  json_struct *document = NULL;
  json_attr_val *avl = NULL;

  json2mcf_check_options(ctx);
  wb = word_buffer_load_mcf(ctx->mcf_filename, ctx->mcd_struct);
  root = json_parse_full(ctx->json_filename, 0);
  
  if(root->type != JSON_AVL){
    fprintf(stderr, "erreur le json doit être un objet\n");
    exit(1);
  }
  for(avl = root->u.attr_val_list; avl != NULL; avl = avl->next){
    printf("section %s\n", avl->attr);
    if(!strcmp(avl->attr, (char *)"documents")){
      process_documents(avl->val, wb);
    }
  }
  
  //json_print_struct(stdout, root);

  json_free_struct(root);
  json2mcf_context_free(ctx);
  return 0;
}
